# pylint: disable=missing-module-docstring, wrong-import-position, import-error, no-member, no-name-in-module, unnecessary-pass, too-few-public-methods
import sys
import os

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from bad.action_network import ActionNetwork
from bad.encoding.observation import Observation
from bad.bayesian_action import BayesianAction

class Policy:
    '''policy'''
    def __init__(self, network: ActionNetwork) -> None:
        '''init'''
        self.network = network
    def get_action(self, observation: Observation) -> BayesianAction:
        '''get action'''
        return self.network.get_action(observation)



# pylint: disable=missing-module-docstring, wrong-import-position, import-error, no-member, no-name-in-module, too-many-function-args, unnecessary-pass, too-few-public-methods
import sys
import os

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from bad.run_episode_result import RunEpisodeResult


class PrintEpisodeSelfPlay:
    '''print episode self play'''
    def __init__(self, run_episode_result: RunEpisodeResult) -> None:
        '''init'''
        self.run_episode_result = run_episode_result

    def print(self) -> None:
        '''print'''
        print(f"episode {self.run_episode_result.number} result:")
        print(f"episode reward: {self.run_episode_result.reward}")
        #print(f"finish: {self.run_episode_result.hanabi_environment.state}")
        #print(f"score: {self.run_episode_result.hanabi_environment.state.score}")
        print(f"life tokens: {self.run_episode_result.hanabi_environment.state.life_tokens}")
        total_steps = len(self.run_episode_result.agent_step_times)
        total_time = self.run_episode_result.agent_step_times.sum()
        print(f"Average training steps per second: {total_steps / total_time}")



# pylint: disable=missing-module-docstring, wrong-import-position, import-error, no-member, no-name-in-module, too-few-public-methods unused-variable
import sys
import os

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from hanabi_learning_environment import pyhanabi, rl_env
from bad.action_network import ActionNetwork
from bad.encoding.observationconverter import ObservationConverter
from bad.set_extra_observation import SetExtraObservation

class TrainEpoch:
    '''train epoch'''
    def __init__(self) -> None:
        players:int = 2
        self.hanabi_environment = rl_env.make('Hanabi-Full', players, \
            pyhanabi.AgentObservationType.CARD_KNOWLEDGE.SEER)

    def train(self, batch_size: int) -> ActionNetwork:
        '''train within an environment'''
        hanabi_observation = self.hanabi_environment.reset()
        # one more move because of no-action move on the beginning
        #fake an action that does not exists
        max_moves: int = self.hanabi_environment.game.max_moves() + 1
        max_actions = max_moves + 1 # 0 index based

        seo = SetExtraObservation()
        seo.set_extra_observation(hanabi_observation, max_moves, max_actions, \
             self.hanabi_environment.state.legal_moves_int())

        observation_converter: ObservationConverter = ObservationConverter()
        observation = observation_converter.convert(hanabi_observation)

        network: ActionNetwork = ActionNetwork()
        network.build(observation, max_actions)

        done = False
        while not done:

            for batch in range(batch_size):
                # legal_actions = self.hanabi_environment.state.legal_moves()
                bad = network.train_observation(observation)
                next_action = bad.decode_action(self.hanabi_environment.state.legal_moves_int())
                next_move = self.hanabi_environment.game.get_move(next_action)

                observation_after_step, _, done, _ = self.hanabi_environment.step(next_action)

                seo.set_extra_observation(observation_after_step, next_action, max_actions, \
                    self.hanabi_environment.state.legal_moves_int())

                observation = observation_converter.convert(observation_after_step)

        return network



# pylint: disable=missing-module-docstring, wrong-import-position, import-error, no-member, no-name-in-module, too-many-function-args
import random
import sys
import os
import numpy as np
import tensorflow as tf

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from bad.run_eposiode import RunEpisode
from bad.train_epoch import TrainEpoch
from bad.print_episode_selfplay import PrintEpisodeSelfPlay
from bad.print_total_selfplay import PrintTotalSelfPlay

class Runner:
    '''runner'''
    def __init__(self) -> None:
        seed = 42
        tf.random.set_seed(seed)
        np.random.seed(seed)
        random.seed(seed)
        self.network = None

    def train(self, episodes: int, batch_size:int) -> None:
        '''train'''
        print('train')
        for episode in range(episodes):
            print(f"begin training for {episode}")
            train_epoch = TrainEpoch()
            self.network = train_epoch.train(batch_size)

    def self_play(self, episodes: int) -> None:
        '''self play'''
        print('self play')

        total_reward = 0
        max_reward = 0
        perfect_games = 0

        for episode in range(episodes):

            run_episode = RunEpisode(self.network)
            episode_result = run_episode.run(episode)

            if episode_result.reward > max_reward:
                max_reward = episode_result.reward
            total_reward += episode_result.reward

            if episode_result.reward == 25:
                perfect_games = perfect_games+1

            print_selfplay = PrintEpisodeSelfPlay(episode_result)
            print_selfplay.print()

        print('')
        print_total_selfplay = PrintTotalSelfPlay(episodes, total_reward, max_reward, perfect_games)
        print_total_selfplay.print()



# pylint: disable=missing-module-docstring, wrong-import-position, import-error
import sys
import os

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from hanabi_learning_environment.rl_env import Agent, HanabiEnv
from bad.bad_agent_acting_result import BadAgentActingResult
from bad.policy import Policy
from bad.encoding.observationconverter import ObservationConverter

class BadAgent(Agent):
    ''' bad agent '''
    def __init__(self, policy: Policy, hanabi_environment: HanabiEnv) -> None:
        self.policy = policy
        self.hanabi_environment = hanabi_environment
        self.observation_converter: ObservationConverter = ObservationConverter()

    def act(self, observation) -> BadAgentActingResult:
        '''act'''
        bad = self.policy.get_action(self.observation_converter.convert(observation))
        next_action = bad.decode_action(self.hanabi_environment.state.legal_moves_int())
        observation_after_step, reward, done, _ = self.hanabi_environment.step(next_action)
        return BadAgentActingResult(observation_after_step, done, int(reward))

    def reset(self, config):
        print('reset')



# pylint: disable=missing-module-docstring, wrong-import-position, import-error, no-member, no-name-in-module, too-few-public-methods unused-variable, no-method-argument, unnecessary-pass, consider-using-enumerate
import sys
import os

import numpy as np

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from hanabi_learning_environment.rl_env import HanabiEnv

class RunEpisodeResult:
    '''run episode result'''
    def __init__(self,number:int, reward:int, hanabi_environment: HanabiEnv, \
        agent_step_times: np.ndarray) -> None:
        self.number = number
        self.reward  = reward
        self.hanabi_environment = hanabi_environment
        self.agent_step_times = agent_step_times



# pylint: disable=missing-module-docstring, wrong-import-position, import-error, too-few-public-methods too-many-arguments, disable=too-many-instance-attributes
import sys
import os
import numpy as np
import tensorflow as tf

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from bad.encoding.fireworkrank import FireworkRank
from bad.encoding.legal_actions import LegalActions

class PublicFeatures:
    '''public features'''
    def __init__(self, observation: dict) -> None:
        '''init'''
        self.observation = observation
        self.curr_player = observation['current_player']

        self.life_tokens_left = self.convert_life_tokens()
        self.hint_tokens_left = self.convert_information_tokens()
        self.current_player = self.convert_current_player()
        self.last_action = self.convert_last_action()
        self.firework = FireworkRank(observation)
        self.legal_actions = LegalActions(observation)

    def convert_life_tokens(self) -> np.ndarray:
        ''' convert life tokens '''
        life_tokens = self.observation['player_observations'][self.curr_player]['life_tokens']
        return tf.keras.utils.to_categorical(life_tokens, num_classes=4, dtype=int)

    def convert_information_tokens(self) -> np.ndarray:
        ''' convert information tokens '''
        in_to = self.observation['player_observations'][self.curr_player]['information_tokens']
        return tf.keras.utils.to_categorical(in_to, num_classes=9, dtype=int)

    def convert_current_player(self) -> np.ndarray:
        '''converts current player'''
        return tf.keras.utils.to_categorical(self.curr_player, num_classes=2, dtype=int)

    def convert_last_action(self) -> np.ndarray:
        '''convert last action'''
        max_action: int = self.observation['max_action']
        last_action: int = self.observation['last_action']

        return tf.keras.utils.to_categorical(last_action, num_classes=max_action, dtype=int)




# pylint: disable=missing-module-docstring, wrong-import-position, import-error
import unittest
import sys
import os

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from bad.nextplayer import NextPlayer


class TestNextPlayer(unittest.TestCase):
    '''test next player'''
    def test_next_player_is_one(self):
        ''' test next player '''
        observation = {}
        observation['current_player'] = 0

        next_player = NextPlayer()
        next_player_id = next_player.next_player(observation)

        self.assertEqual(next_player_id, 1)

    def test_next_player_is_zero(self):
        ''' test next player '''
        observation = {}
        observation['current_player'] = 1

        next_player = NextPlayer()
        next_player_id = next_player.next_player(observation)

        self.assertEqual(next_player_id, 0)



# pylint: disable=missing-module-docstring, wrong-import-position, import-error too-few-public-methods too-many-arguments

class NextPlayer:
    '''next player (may should belong to hanabi environment itself)'''
    def next_player(self, observation: dict) -> int:
        '''next player'''
        curr_player:int = observation['current_player']
        return (curr_player +1) % 2



# pylint: disable=missing-module-docstring, wrong-import-position, import-error
import os
import sys
import tensorflow as tf

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from bad.runner import Runner

def main() -> None:
    '''main'''
    batch_size: int = 1
    episodes_training = 1
    episodes_running = 100

    print(f'welcome to bad agent with tf version: {tf.__version__}')
    print(f'running {episodes_running} episodes')

    runner = Runner()
    runner.train(episodes=episodes_training, batch_size=batch_size)
    runner.self_play(episodes_running)

    print("finish with everything")
if __name__ == "__main__":
    main()



# pylint: disable=missing-module-docstring, wrong-import-position, import-error too-few-public-methods too-many-arguments
import numpy as np
import tensorflow as tf


class RankToIntConverter:
    ''' rank converter'''
    def convert(self, rank:int) -> np.ndarray:
        '''convert'''
        return tf.keras.utils.to_categorical(rank, num_classes=6, dtype=int)
